// This file must be the first include for any bindings code.

#pragma once

#include <pybind11/pybind11.h>

namespace mqt {
namespace py = pybind11;
using namespace py::literals;

} // namespace mqt
